# coding=UTF-8

import os

from PyQt5.QtWidgets import QDialog
from PyQt5 import uic

from qgis.core import QgsProject, QgsMapLayer

from .config import PlotConfig


class ConfigCreateDialog(QDialog):

    def __init__(self, parent, timeseries=False):
        """ ConfigCreateDialog is used to add a new log configuration

        :param timeseries: true if times series log
        :param timeseries: bool"""
        super().__init__(parent)

        uic.loadUi(os.path.join(os.path.dirname(__file__),
                                'config_create_dialog.ui'), self)
        self.timeseries = timeseries
        if timeseries:
            types=['View measures instantaneous', 'View measures cumulative']
        else:
            types=['Stratigraphic', 'Log measures cumulative','Log measures continuous']
        self._type.addItems(types)
        self._type.currentIndexChanged.connect(self.__on_type_changed)
        self.__on_type_changed(0)

        self._source.currentIndexChanged.connect(self._on_source_changed)
        for layer in QgsProject.instance().mapLayers().values():
            if layer.type() == QgsMapLayer.VectorLayer:
                self._source.addItem(layer.name(), layer)

        style_dir = os.path.join(os.path.dirname(__file__),'qgeologis','styles')
        self._style_file.addItems([f for f in os.listdir(style_dir) if f[-4:]=='.xml'])

    def __on_type_changed(self, index):
        """ Triggered on log type combobox change

        :param index: log type combobox index
        :type index: int
        """
        if index == 0 and self.timeseries :
            index = 3
        self._type_widgets.setCurrentIndex(index)

    def _on_source_changed(self, index):
        """ Triggered on type change, fill form with layer field

        :param index: page index
        :type index: int
        """
        layer = self._source.itemData(index)

        for column_field in [self._feature_ref_column, self._depth_from_column,
                             self._depth_to_column, self._start_measure_column,
                             self._interval_column, self._value_continuous_column,
                             self._event_column, self._value_instantaneous_column,
                             self._start_cumulative_column, self._end_cumulative_column,
                             self._value_cumulative_column]:
            column_field.clear()
            column_field.addItems(layer.fields().names())

        for column_field in [self._formation_code_column, self._rock_code_column,
                             self._formation_description_column, self._rock_description_column]:
            column_field.clear()
            column_field.addItems([''] + layer.fields().names())

    def config(self):
        """ Return configuration from dialog

        :return: configuration
        :rtype: dict
        """
        # Stratigraphy
        config = None
        if self._type.currentIndex() == 0 and self.timeseries == False :
            config = {
                "type": 'stratigraphic',
                "depth_from_column": self._depth_from_column.currentText(),
                "depth_to_column": self._depth_to_column.currentText(),
                "formation_code_column": self._formation_code_column.currentText() or None,
                "rock_code_column": self._rock_code_column.currentText() or None,
                "formation_description_column": self._formation_description_column.currentText() or None,
                "rock_description_column": self._rock_description_column.currentText() or None,
                "style": self._style_file.currentText(),
                "display": "1"
                }

        # log measure instantaneous
        elif self._type.currentIndex() == 0 and self.timeseries == True :
            config = {
                "start_measure_column": self._start_measure_column.currentText(),
                "event_column": self._event_column.currentText(),
                "value_column": self._value_instantaneous_column.currentText(),
                "uom":  self._uom_instantaneous.text(),
                "type": "instantaneous",
                "display": "1"
                }

        # log measure cumulative
        elif self._type.currentIndex() == 1:
            config = {
                "uom": self._uom_cumulative.text(),
                "type": 'cumulative',
                "min_event_column": self._start_cumulative_column.currentText(),
                "max_event_column": self._end_cumulative_column.currentText(),
                "value_column": self._value_cumulative_column.currentText(),
                "display": "1"
                }

        # log measure continuous
        elif self._type.currentIndex() == 2:
            config = {
                "start_measure_column": self._start_measure_column.currentText(),
                "interval_column": self._interval_column.currentText(),
                "values_column": self._value_continuous_column.currentText(),
                "uom":  self._uom_continuous.text(),
                "type": "continuous",
                "min": self._continuous_min.text() if self._continuous_min.isEnabled() else None,
                "max": self._continuous_max.text() if self._continuous_max.isEnabled() else None,
                "display": "1"
                }

        config["source"] = self._source.itemData(self._source.currentIndex()).id()
        config["feature_ref_column"] = self._feature_ref_column.currentText()
        config["name"] = self._name.text()

        if self.timeseries :
            section = "timeseries"
        else :
            section = ["stratigraphy_config", "log_measures", "log_measures"][self._type.currentIndex()]
        return (section, PlotConfig(config))